
############################################################################################################

# Real time Deeplabcut  
# Ali Nourizonoz & Yannick Ormen 
# ali.nourizonoz@unige.ch - ormeny@ethz.ch
# Macrh 2019
# Huberlab | huberlab.org
# Department of Basic Neurosciences  
# University of Geneva




############################################################################################################
#     importing all depedencies needed for realtime Deeplabcut
############################################################################################################

from deeplabcut.pose_estimation_tensorflow.nnet import predict
from deeplabcut.pose_estimation_tensorflow.config import load_config
from deeplabcut.pose_estimation_tensorflow.dataset.pose_dataset import data_to_input
from deeplabcut.pose_estimation_tensorflow.nnet.net_factory import pose_net
import time
import pandas as pd
import numpy as np
import os
import argparse
from pathlib import Path
from tqdm import tqdm
import tensorflow as tf
from deeplabcut.utils import auxiliaryfunctions
import cv2
from skimage.util import img_as_ubyte
from pylab import * 
import PIL

from ximea import xiapi
import sys
import socket
import pickle
import itertools
import imageio
import skimage
import skimage.color
import glob
from threading import Thread

config = tf.ConfigProto()
config.gpu_options.allow_growth=True
sess = tf.Session(config=config)

############################################################################################################
#     Path configuration and setups for initial weights 
############################################################################################################

config_path = '/home/huberlab/DeepLabCut/mouse_ali-ormen-2019-08-02/config.yaml'   # This is your config file path, you have to change this for your own config file path
start_path=os.getcwd()
cfg = auxiliaryfunctions.read_config(config_path)
trainFraction = cfg['TrainingFraction'][0]
modelfolder=os.path.join(cfg["project_path"],str(auxiliaryfunctions.GetModelFolder(trainFraction,1,cfg)))
path_test_config = Path(modelfolder) / 'test' / 'pose_cfg.yaml'
dlc_cfg = load_config(str(path_test_config))
Snapshots = np.array([fn.split('.')[0]for fn in os.listdir(os.path.join(modelfolder , 'train'))if "index" in fn])

if cfg['snapshotindex'] == 'all':
    print("Snapshotindex is set to 'all' in the config.yaml file. Running video analysis with all snapshots is very costly! Use the function 'evaluate_network' to choose the best the snapshot. For now, changing snapshot index to -1!")
    snapshotindex = -1
else:
    snapshotindex=cfg['snapshotindex']
		
increasing_indices = np.argsort([int(m.split('-')[1]) for m in Snapshots])
Snapshots = Snapshots[increasing_indices]
dlc_cfg['init_weights'] = os.path.join(modelfolder , 'train', Snapshots[snapshotindex])
trainingsiterations = (dlc_cfg['init_weights'].split(os.sep)[-1]).split('-')[-1]
sess, inputs, outputs = predict.setup_pose_prediction(dlc_cfg)



############################################################################################################
#     Core function for pose extraction from captured images (labeling)
############################################################################################################

# This function is taken from the "" file in Deeplabcut
def getpose(image,sess, inputs, cfg, outputs, outall=False):
    ''' Adapted from DeeperCut, see pose-tensorflow folder'''
  
    im=np.expand_dims(skimage.color.gray2rgb(image), axis=0).astype(float)
    outputs_np = sess.run(outputs, feed_dict={inputs: im})
    scmap, locref = predict.extract_cnn_output(outputs_np, cfg)
    pose = predict.argmax_pose_predict(scmap, locref, cfg.stride)
    if outall:
        return scmap, locref, pose
    else:
        return pose					

#the pose is basically Nx3.  N rows (body parts as listed in config file). 
#The 3 columns are   (Probability of detection,  X_position,  Y_position).


############################################################################################################
#     Creating UDP socket for sending all body parts in realtime to another machine
############################################################################################################

serverAddressPort = ("10.0.0.100",5555)			# IP adr and port for the host machine 
bufferSize =1024
UDPClientSocket = socket.socket(family=socket.AF_INET,type=socket.SOCK_DGRAM) 




############################################################################################################
#    All important variables used
############################################################################################################

pcutoff = cfg['pcutoff']					# Cutoff for pose detection (cutoff for likelihood of each body part) in range of 0-10
t_end=60							# Recording time in seconds 
FrameNum=0							# Frame number
						# Starting time 
bodyparts2plot = cfg['bodyparts']				# Make a list of all body parts to be tracked
colors = plt.cm.get_cmap(cfg['colormap'],len(bodyparts2plot)) 	# Generate N (number of body parts) distinc colors for labling image
print("Starting to extract posture")

############################################################################################################
#    Loop for realtime Deeplabcut
############################################################################################################
data=np.zeros(shape=[640,512,1],dtype=np.uint8) 
con=0
cc=0

############################################################################################################
#    Oppening XIMEA camera and setting up camera parameters for fast image acquisition
############################################################################################################



def myfunc():
   global data 
   global con
   global cc
   cam = xiapi.Camera()
   cam.open_device()
   cam.disable_aeag()
   cam.set_imgdataformat('XI_RAW8')  			
   cam.set_exposure(6000) 
   cam.set_gain(20)					
   cam.set_downsampling_type('XI_SKIPPING') 		
   cam.set_downsampling('XI_DWN_2x2') 			
   img = xiapi.Image()					
   cam.start_acquisition()
   
   while(1):
       cam.get_image(img)
       data = img.get_image_data_numpy()	
       con=1
       cc=cc+1
      # print(cc)



t = Thread(target=myfunc)
t.start()

time.sleep(5)
start = time.time()
mycc=0
a = datetime.datetime.now()
textfile = open('/home/huberlab/Desktop/textfile.txt', 'w')
while (time.time()-start) < t_end:
   
    if cc>mycc :
        mycc=cc
        
        data2 = cv2.resize(data,(320,256)) 			# Resize image for getpose function. Here we do it 2x2. This will increase the speed.. 
  

   					
    
    #Passing the image to getpose function for labeling

        pose = getpose(data2,sess, inputs, dlc_cfg, outputs)			
        pose[:,0]=pose[:,0]*2  				# X position is multiplied by 2 (since we already resized the image to 320x256)
        pose[:,1]=pose[:,1]*2				# Same for Y position
        pose[:,2]=pose[:,2]*10   				# The probability multiplied by 10 (0-10)
    

    # Sending all probabilities and body coordinate as a string via UDP to another machine

        pp=pose.reshape(-1)						# Sort all elements of pose in a single row structure
        st=str(FrameNum)+" "+" ".join(str((x)) for x in pp)	#string starting with the frame number followed by X,Y and Probability (P) of each body part (FrameNum, X1,Y1,P1 ,X2,Y2,P2 ..)
        
      	# YOU CAN SAVE THIS STRING IN A FILE TO HAVE COORDINATES AND LIKELIHOODS LATER
        #print(st)
        
        UDPClientSocket.sendto(st.encode('utf-8'),serverAddressPort) # Sending the string of body coordinates with their liklihood as a UDP packet

	################################################################################################FILTER DESIGNE##############################################################
	# You can designe your filter here!!!!!!!!!!!!!!!!!!!!!!!!!
	#	if(pos[1,1]-pos[2,1]>200 & pos[1,1]-pos[2,1]<200)
	#		st=str(FrameNum)+" "+str(1)
	#	else
	#		st=str(FrameNum)+ " "str(0)
	#	UDPClientSocket.sendto(st.encode('utf-8'),serverAddressPort)




    # Show the image and tracked body parts (as colored dots)
       # print(pcutoff)
        #img2=cv2.cvtColor(data,cv2.COLOR_GRAY2RGB)		# Image from camera is RAW8(gray scale) so to plot colored dot we convert the image to 3 channel RGB image
        #for i in range(0,len(bodyparts2plot)):             
          # if pose[i,2] > pcutoff:  			# For each body part, plot a distinc color (from our color map described earlier) if the it's probability is over pcutoff
           #   cv2.circle(img2,(int(pose[i,0]),int(pose[i,1])),3,[x*255 for x in colors(i)],-1)
    
        #cv2.imshow('img',img2)				# Display the image
        #cv2.waitKey(1)
    

    # Saving the original  image to file with format IMG_framenumber

        filename="/home/huberlab/Desktop/M6/IMG_%d.jpg"%FrameNum;
        cv2.imwrite(filename,data) 
    
        FrameNum+=1
        b = datetime.datetime.now()
        c = b - a
        st=st+" "+str(c.total_seconds())+"\n"
        textfile.write(st)
        #print(c.seconds*1000000+c.microseconds)
        #print(c.total_seconds())




stop = time.time()
con=-1
print(FrameNum)
print(cc)


